/*
 *     gtkatlantic - the gtk+ monopd client, enjoy network monopoly games
 *
 *
 *  Copyright (C) 2002-2003 Rochet Sylvain
 *
 *  gtkatlantic is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ** functions for connect/disconnect server, receive/send data to/from server */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>

#include "config.h"
#include "xmlparse.h"
#include "global.h"
#include "client.h"
#include "game.h"


/* -- connect to server
     >
     > return false if:
     >   - max number of connection reached
     >   - can't resolve host
     >   - can't open port or socket
*/
gboolean client_connect(guint16 *idconnect, gchar* host, guint32 port)  {

	struct sockaddr_in addr;
	struct hostent *ent;
	struct in_addr in;
	gint32 addr_len, test;
	gpointer errormsg;
	gchar *text;
	guint16 idcon;
	guint32 sock;


	if( !connect_get_valid_id(&idcon) )  {

		errormsg = g_strdup("[ERROR] Max number of connections reached");
		text_insert_message(errormsg, strlen(errormsg));
		g_free(errormsg);
		return(FALSE);
	}

	/** resolve an address */   
	if (!inet_aton(host, &in) ) {

		if (! (ent = gethostbyname(host)) )  {

			errormsg = g_strdup_printf("[ERROR] Can't resolve host \"%s\"", host);
			text_insert_message(errormsg, strlen(errormsg));
			g_free(errormsg);
			return(FALSE);

		}  else  {
      
			in = *(struct in_addr *)ent->h_addr;
			addr.sin_family = ent->h_addrtype;   
		}
	}
	else  addr.sin_family = AF_INET;

	addr.sin_addr = in;
	addr.sin_port = htons(port);
	addr_len = sizeof(addr);   

	/* -- now host is resolved, continue connection */
	sock = socket(addr.sin_family, SOCK_STREAM, IPPROTO_IP);
		// see /usr/include/netinet/in.h for IPPROTO_x

	test = connect(sock, (struct sockaddr*) &addr, addr_len);

	if(test < 0) {

		errormsg = g_strdup_printf("[ERROR] Host \"%s\" resolved but can't open port \"%d\" or socket", host, port);
		text_insert_message(errormsg, strlen(errormsg));
		g_free(errormsg);
		return(FALSE);
	}

	/* -- connection established :), finishing it */
	connection[idcon] = g_malloc0(sizeof (_connect) );
	connection_is_open[idcon] = TRUE;

	connection[idcon]->socket = sock;
	connection[idcon]->host = g_strdup(host);
	connection[idcon]->port = port;
	connection[idcon]->connected = TRUE;
	*idconnect = idcon;

	text = g_strdup_printf("[CONNECT %d] connected on \"%s:%d\", use socket %d", idcon, host, port, sock);
	text_insert_message(text, strlen(text));
	g_free(text);

	return(TRUE);
}


/* -- close the connection */
gboolean client_disconnect(guint32 idconnect)  {

	gpointer errormsg;
	gchar *text;

	if(!connection_is_open[idconnect])   return(FALSE);
	if(!connection[idconnect]->connected)   return(FALSE);

	shutdown (connection[idconnect]->socket, 2);

	if(close(connection[idconnect]->socket))  {

		errormsg = g_strdup_printf("[ERROR] Connection \"%d\" : can't close socket", idconnect);
		text_insert_message(errormsg, strlen(errormsg));
		g_free(errormsg);
		return(FALSE);
	}

	text = g_strdup_printf("[CONNECT %d] socket %d closed", idconnect, connection[idconnect]->socket);
	text_insert_message(text, strlen(text));
	g_free(text);

	if(connection[idconnect]->lenght_received > 0)
		g_free(connection[idconnect]->buf_received);

	g_free(connection[idconnect]->host);
	g_free(connection[idconnect]->server_version);

	g_free( connection[idconnect] );
	connection_is_open[idconnect] = FALSE;

	return(TRUE);
}


/* -- exec all client fonctions, for all games */
gboolean client_fonct()  {

	guint32 i = 0;

	for(i = 0 ; i < MAX_CONNECTION; i++) {

		if(connection_is_open[i])
			if(connection[i]->connected)
				while( client_recv(i) )  {
          
					if(connection[i]->line_finished)
						switch(connection[i]->type)  {

							case CONNECT_TYPE_MONOPD_GETGAME:
								xmlparse_getgamelist_plugger(i);
								break;
							case CONNECT_TYPE_MONOPD_GAME:
								xmlparse_game_plugger(i);
								break;

							case CONNECT_TYPE_META_GETLIST:
							case CONNECT_TYPE_META_GETGAME:
								xmlparse_metaserver(i);
								break;
						}
				}
	}

	return(TRUE);
}


/* -- get data from server
      >> store it in get_tmp buffer */
gboolean client_recv(guint32 id)  {

	guint32 lenght_data, i, sock;
	guchar *data = 0, *tmp;  /* tmp ptr */
	gpointer errormsg;

	if(!connection_is_open[id])  return(FALSE);
	if(!connection[id]->connected)  return(FALSE);
	sock = connection[id]->socket;
	connection[id]->line_finished = FALSE;

	for(i = 0 ; client_can_get_data_from_server(id) ; i++) {
   
		if(!data)  data = g_malloc0(GETDATA_STEP +SECURITY_SIZE);

		if(! (i % GETDATA_STEP))  {
      
			data = g_realloc(data, i + GETDATA_STEP +SECURITY_SIZE);  
			memset(data + i, 0, GETDATA_STEP +SECURITY_SIZE);
		}

		if( read(sock, &data[i], 1) != 1) {

			errormsg = g_strdup("[ERROR] connection to server lost :(");
			text_insert_message(errormsg, strlen(errormsg));
			g_free(errormsg);

			client_disconnect(game->connectid);
			g_free(data);
			return(FALSE);
		}

		if( data[i] == '\n')  {  /* read it one line at a time */

			connection[id]->line_finished = TRUE;
			break;
		}
	}

	/* quit if no data from server */
	if(!data)  return(FALSE);

	lenght_data = ++i;
	data[i] = '\0';

	/* replace undesired char */
	for(i = 0; i < lenght_data ; i++)
		if( data[i] > 127 )  {

			switch(data[i])  {

				case 161:  data[i] = '!';  break;
				case 162:  data[i] = 'c';  break;
				case 163:  data[i] = 'f';  break;
				case 164:  data[i] = 'o';  break;
				case 165:  data[i] = 'Y';  break;
				case 166:  data[i] = '|';  break;
				case 167:  data[i] = 'S';  break;
				case 169:  data[i] = 'c';  break;
				case 170:  data[i] = 'a';  break;
				case 171:  data[i] = 'C';  break;
				case 172:  data[i] = '-';  break;
				case 173:  data[i] = '-';  break;
				case 174:  data[i] = 'R';  break;
				case 175:  data[i] = '-';  break;
				case 176:  data[i] = 'o';  break;
				case 177:  data[i] = '+';  break;
				case 178:  data[i] = '2';  break;
				case 179:  data[i] = '3';  break;
				case 181:  data[i] = 'u';  break;
				case 182:  data[i] = 'P';  break;
				case 183:  data[i] = '-';  break;
				case 185:  data[i] = '1';  break;
				case 186:  data[i] = 'o';  break;
				case 187:  data[i] = 'D';  break;
				case 191:  data[i] = '?';  break;
				case 192:  data[i] = 'A';  break;
				case 193:  data[i] = 'A';  break;
				case 194:  data[i] = 'A';  break;
				case 195:  data[i] = 'A';  break;
				case 196:  data[i] = 'A';  break;
				case 197:  data[i] = 'A';  break;
				case 198:  data[i] = 'F';  break;
				case 199:  data[i] = 'C';  break;
				case 200:  data[i] = 'E';  break;
				case 201:  data[i] = 'E';  break;
				case 202:  data[i] = 'E';  break;
				case 203:  data[i] = 'E';  break;
				case 204:  data[i] = 'I';  break;
				case 205:  data[i] = 'I';  break;
				case 206:  data[i] = 'I';  break;
				case 207:  data[i] = 'I';  break;
				case 208:  data[i] = 'D';  break;
				case 209:  data[i] = 'N';  break;
				case 210:  data[i] = 'O';  break;
				case 211:  data[i] = 'O';  break;
				case 212:  data[i] = 'O';  break;
				case 213:  data[i] = 'O';  break;
				case 214:  data[i] = 'O';  break;
				case 215:  data[i] = 'x';  break;
				case 216:  data[i] = 'O';  break;
				case 217:  data[i] = 'U';  break;
				case 218:  data[i] = 'U';  break;
				case 219:  data[i] = 'U';  break;
				case 220:  data[i] = 'U';  break;
				case 221:  data[i] = 'Y';  break;
				case 222:  data[i] = 'p';  break;
				case 223:  data[i] = 'B';  break;
				case 224:  data[i] = 'a';  break;
				case 225:  data[i] = 'a';  break;
				case 226:  data[i] = 'a';  break;
				case 227:  data[i] = 'a';  break;
				case 228:  data[i] = 'a';  break;
				case 229:  data[i] = 'a';  break;
				case 230:  data[i] = 'e';  break;
				case 231:  data[i] = 'c';  break;
				case 232:  data[i] = 'e';  break;
				case 233:  data[i] = 'e';  break;
				case 234:  data[i] = 'e';  break;
				case 235:  data[i] = 'e';  break;
				case 236:  data[i] = 'i';  break;
				case 237:  data[i] = 'i';  break;
				case 238:  data[i] = 'i';  break;
				case 239:  data[i] = 'i';  break;
				case 240:  data[i] = 'o';  break;
				case 241:  data[i] = 'n';  break;
				case 242:  data[i] = 'o';  break;
				case 243:  data[i] = 'o';  break;
				case 244:  data[i] = 'o';  break;
				case 245:  data[i] = 'o';  break;
				case 246:  data[i] = 'o';  break;
				case 247:  data[i] = '%';  break;
				case 248:  data[i] = 'o';  break;
				case 249:  data[i] = 'u';  break;
				case 250:  data[i] = 'u';  break;
				case 251:  data[i] = 'u';  break;
				case 252:  data[i] = 'u';  break;
				case 253:  data[i] = 'y';  break;
				case 254:  data[i] = 'p';  break;
				case 255:  data[i] = 'y';  break;
				default:
#if DEBUG
					fprintf(stderr, "[WARNING] No UTF8 chr from server: (%c -> %d)\n", data[i], data[i]);
#endif
					data[i] = ' ';
					break;
			}
		}

	/* copy data */
	if(!connection[id]->lenght_received)  {

		connection[id]->buf_received = g_strdup(data);
	}
	else   {

		tmp = connection[id]->buf_received;
		connection[id]->buf_received = g_strconcat(connection[id]->buf_received, data, NULL);
		g_free(tmp);
	}

	connection[id]->lenght_received += lenght_data;

#if DEBUG
	fprintf(stdout, "\033[32mID(%.2d): GET(%.4d): [%s]\033[m\n", id, lenght_data, data);
//	fprintf(stdout, "\033[36mID(%.2d): GET(%.4d): [%s]\033[m\n", id, connection[id]->lenght_received, connection[id]->buf_received);
#endif

	g_free(data);
	return(TRUE);
}


/* return TRUE if there are data from server
   else return FALSE */
gboolean client_can_get_data_from_server(guint32 id)  {

	fd_set rfds;
	struct timeval tv;
	guint32 sock;

	sock = connection[id]->socket;
   
	FD_ZERO (&rfds);
	FD_SET (sock, &rfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	if( select (sock + 1, &rfds, NULL, NULL, &tv) ) return(TRUE);

	return(FALSE);
}


/* -- send data to server
      >> get from send_tmp buffer
      >> flush send_tmp buffer */
gboolean client_send(guint32 id, gchar *data, guint32 lenght_data)  {

	guint32 t;
	gpointer errormsg;

	if(!connection_is_open[id])  return(TRUE);
	if(!connection[id]->connected)  return(TRUE);

	t = write(connection[id]->socket, data, lenght_data);

	if(t != lenght_data) {

		errormsg = g_strdup("[ERROR] Can't send data to server");
		text_insert_message(errormsg, strlen(errormsg));
		g_free(errormsg);
		return(FALSE);
	}

#if DEBUG
	fprintf(stdout, "\033[31mID(%.2d): SEND(%.4d): [%s]\033[m\n", id, lenght_data, data);
#endif

	return(TRUE);
}


